/*
===========================================================================

Doom 3 GPL Source Code
Copyright (C) 1999-2011 id Software LLC, a ZeniMax Media company.

This file is part of the Doom 3 GPL Source Code (?Doom 3 Source Code?).

Doom 3 Source Code is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Doom 3 Source Code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Doom 3 Source Code.  If not, see <http://www.gnu.org/licenses/>.

In addition, the Doom 3 Source Code is also subject to certain additional terms. You should have received a copy of these additional terms immediately following the terms and conditions of the GNU General Public License which accompanied the Doom 3 Source Code.  If not, please request a copy in writing from id Software at the address below.

If you have questions concerning this license or the applicable additional terms, you may contact in writing id Software LLC, c/o ZeniMax Media Inc., Suite 120, Rockville, Maryland 20850 USA.

===========================================================================
*/

#ifndef __PHYSICS_ACTOR_H__
#define __PHYSICS_ACTOR_H__

/*
===================================================================================

	Actor physics base class

	An actor typically uses one collision model which is aligned with the gravity
	direction. The collision model is usually a simple box with the origin at the
	bottom center.

===================================================================================
*/

class idPhysics_Actor : public idPhysics_Base
{

public:
    CLASS_PROTOTYPE(idPhysics_Actor);

    idPhysics_Actor(void);
    ~idPhysics_Actor(void);

    void					Save(idSaveGame *savefile) const;
    void					Restore(idRestoreGame *savefile);

    // get delta yaw of master
    float					GetMasterDeltaYaw(void) const;
    // returns the ground entity
    idEntity 				*GetGroundEntity(void) const;
    // align the clip model with the gravity direction
    void					SetClipModelAxis(void);

public:	// common physics interface
    void					SetClipModel(idClipModel *model, float density, int id = 0, bool freeOld = true);
    idClipModel 			*GetClipModel(int id = 0) const;
    int						GetNumClipModels(void) const;

    void					SetMass(float mass, int id = -1);
    float					GetMass(int id = -1) const;

    void					SetContents(int contents, int id = -1);
    int						GetContents(int id = -1) const;

    const idBounds 		&GetBounds(int id = -1) const;
    const idBounds 		&GetAbsBounds(int id = -1) const;

    bool					IsPushable(void) const;

    const idVec3 			&GetOrigin(int id = 0) const;
    const idMat3 			&GetAxis(int id = 0) const;

    void					SetGravity(const idVec3 &newGravity);
    const idMat3 			&GetGravityAxis(void) const;

    void					ClipTranslation(trace_t &results, const idVec3 &translation, const idClipModel *model) const;
    void					ClipRotation(trace_t &results, const idRotation &rotation, const idClipModel *model) const;
    int						ClipContents(const idClipModel *model) const;

    void					DisableClip(void);
    void					EnableClip(void);

    void					UnlinkClip(void);
    void					LinkClip(void);

    bool					EvaluateContacts(void);

protected:
    idClipModel 			*clipModel;			// clip model used for collision detection
    idMat3					clipModelAxis;		// axis of clip model aligned with gravity direction

    // derived properties
    float					mass;
    float					invMass;

    // master
    idEntity 				*masterEntity;
    float					masterYaw;
    float					masterDeltaYaw;

    // results of last evaluate
    idEntityPtr<idEntity>	groundEntityPtr;
};

#endif /* !__PHYSICS_ACTOR_H__ */
